phosh-upcoming-events {
  background-color: @phosh_notification_bg_color;
  border-radius: 12px;
  padding: 8px;
}

phosh-upcoming-events list {
  background-color: transparent;
}

phosh-upcoming-events row {
  border-radius: 6px;
  background: shade(@phosh_notification_bg_color, 1.3);
}

phosh-upcoming-events row:hover {
  background: shade(@phosh_notification_bg_color, 1.5);
}

phosh-upcoming-events row:active {
  background: shade(@phosh_notification_bg_color, 1.8);
}

phosh-upcoming-events phosh-upcoming-event .color-bar {
  border-radius: 6px;
  background-color: currentColor;
}
