/* Adwaita dark theme variant */

@define-color phosh_fg_color white;
@define-color phosh_bg_color black;

@define-color phosh_borders_color alpha(@phosh_fg_color,.1);

@define-color phosh_activity_bg_color #1a1a1a;
@define-color phosh_notification_bg_color #282828;
@define-color phosh_action_bg_color #474747;
@define-color phosh_splash_bg_color #f6f5f4;
@define-color phosh_splash_fg_color #282828;

/* Button colors */
@define-color phosh_button_bg_color #282828;
@define-color phosh_button_hover_bg_color shade(@phosh_button_bg_color, 1.14);
@define-color phosh_button_active_bg_color shade(@phosh_button_bg_color, 1.5);

@define-color phosh_emergency_button_bg_color #e01b24;
@define-color phosh_emergency_button_fg_color #ffffff;

@import url("resource:///sm/puri/phosh/stylesheet/common.css");
